@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 1997 Red Hat Software, Inc.
- * Copyright (C) 2001 Red Hat, Inc.
+ * Copyright (C) 2001, 2003, 2007 Red Hat, Inc.
  *
  * This is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by
@@ -36,6 +36,7 @@
  *   question. 
  */
 
+#include "config.h"
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/types.h>
@@ -50,19 +51,20 @@
 #include <string.h>
 #include <unistd.h>
 #include <wait.h>
+
+#include <blkid/blkid.h>
 #include <glib.h>
+#include <glib/gi18n.h>
 #include <gtk/gtk.h>
-#include <glade/glade.h>
-#include "config.h"
 #include "userdialogs.h"
 
-#define N_(String)	String
-#define _(String)	gettext(String)
 #define MOUNT_TEXT	_("_Mount")
 #define UNMOUNT_TEXT	_("Un_mount")
 #define PREFERRED_FS	"ext2"
-#define ACTION_FORMAT	1
-#define ACTION_MOUNT	2
+enum {
+	ACTION_FORMAT = 1,
+	ACTION_MOUNT = 2,
+};
 #define PAD		8
 
 /* The structure which holds all of the information needed to process a
@@ -74,33 +76,66 @@ struct mountinfo {
 	gboolean mounted;	/* TRUE=mounted, FALSE=unmounted */
 	gboolean writable;	/* TRUE=+w, FALSE=-w */
 	gboolean fdformat;	/* TRUE=should run fdformat, FALSE=can't */
-	GtkWidget *mount;	/* mount button */
-	GtkWidget *format;	/* format button */
-	GtkWidget *mount_label;	/* label in the mount button */
 	struct mountinfo *next;	/* linked list... */
 };
 
 /* The tree view widget which contains the needed mountinfo struct in its
- * fourth column.  We can't pass this in any other way because a GtkDialog
- * response callback doesn't have a userdata argument.  */
+ * fourth column. */
 static GtkTreeSelection *tree_selection = NULL;
 
 /* Static pointer to the dialog's buttons. */
 static GtkWidget *mount_button = NULL, *format_button = NULL;
 
-/* This function parses /etc/fstab using getmntent() and uses other tricks
- * to fill in the info for all filesystems with the 'user' option... unless,
- * of course getuid() == 0, then we get them all.  No reason root shouldn't
- * be allowed to use this, right?.  */
-static struct mountinfo *
-build_mountinfo_list()
+/* Check whether PATH is mounted
+   Return -1 if PATH can't ever possibly be mounted, 0 or 1 otherwise */
+static int
+is_mounted(const char *path)
 {
-	struct mountinfo *ret = NULL;
-	struct mountinfo *tmp = NULL;
 	char *parent_dir;
+	int res;
+	struct stat st1, st2;
+
+	if (stat(path, &st1) != 0)
+		goto err;
+	/* Get the name of the mountpoint's parent directory. */
+	if (strrchr(path, '/') == path)
+		/* The mountpoint is a top-level directory. */
+		parent_dir = g_strdup("/");
+	else
+		/* The parent is not the root directory. */
+		parent_dir = g_strconcat(path, "/..", NULL);
+
+	/* Verify that the mountpoint and its parent exist. */
+	res = stat(parent_dir, &st2);
+	g_free(parent_dir);
+	if (res != 0)
+		goto err;
+
+	/* Check if the filesystem is mounted. */
+	return st1.st_dev != st2.st_dev;
+
+err:
+	return -1;
+}
+
+/* This function parses /etc/fstab using getmntent() and uses other tricks
+ * to fill in the info for all filesystems with the 'pamconsole' option...
+ * unless, of course getuid() == 0, then we get them all.  No reason root
+ * shouldn't be allowed to use this, right?.  */
+static struct mountinfo *
+build_mountinfo_list(void)
+{
+	gboolean superuser;
+	blkid_cache cache;
+	struct mountinfo *ret;
 	struct mntent *fstab_entry;
 	FILE *fstab;
 
+	superuser = (geteuid() == 0);
+
+	cache = NULL;
+	(void)blkid_get_cache(&cache, NULL);
+
 	/* Open the /etc/fstab file (yes, the preprocessor define is named
 	 * with an "m", don't let that throw you. */
 	fstab = setmntent(_PATH_MNTTAB, "r");
@@ -108,68 +143,47 @@ build_mountinfo_list()
 
 	/* Iterate over all of the entries. */
 	while((fstab_entry = getmntent(fstab)) != NULL) {
-		struct stat st, st1, st2;
-		const char *node, *fstype, *mountpoint;
-		char *mkfs, *sbindir;
-		gboolean owner, user, swap, superuser;
+		struct stat dev_st;
+		struct mountinfo *tmp;
+		const char *mountpoint;
+		char *dev, *mkfs, *sbindir;
+		gboolean known_device, owner, usable;
+		int res;
 
-		/* Read flags. */
-		owner = (hasmntopt(fstab_entry, "owner") != NULL);
-		user = (hasmntopt(fstab_entry, "user") != NULL);
-		swap = (strcmp(fstab_entry->mnt_type, "swap") == 0);
-		superuser = (geteuid() == 0);
-		node = fstab_entry->mnt_fsname;
-		fstype = fstab_entry->mnt_type;
-		mountpoint = fstab_entry->mnt_dir;
+		dev = blkid_get_devname(cache, fstab_entry->mnt_fsname, NULL);
+		if (dev == NULL)
+			dev = fstab_entry->mnt_fsname;
+		if (strncmp(dev, "/dev/", 5) == 0 && stat (dev, &dev_st) == 0)
+			known_device = TRUE;
+		else
+			known_device = FALSE;
+		if (dev != fstab_entry->mnt_fsname)
+			free(dev);
 
-		/* If the device is in /dev, and the current user owns it,
-		 * make note of the fact that we own it. */
-		memset(&st, 0, sizeof(st));
-		if(owner && (strncmp(node, "/dev/", 5) == 0)) {
-			owner = FALSE;
-			if(stat(node, &st) != -1) {
-				if(st.st_uid == getuid()) {
-					owner = TRUE;
-				}
-			}
-		}
+		owner = (hasmntopt(fstab_entry, "owner") != NULL
+			 && known_device && dev_st.st_uid == getuid());
 
 		/* We only process this entry if we are allowed to mount it
-		 * (it has the "user" flag, or it has the "owner" flag and
-		 * we own the device), and it's not a swap partition. */
-		if((!user && !owner && !superuser) || swap) {
+		   and it's not a swap partition. */
+		usable = (superuser
+			  || hasmntopt(fstab_entry, "pamconsole") != NULL
+			  || hasmntopt(fstab_entry, "user") != NULL
+			  || hasmntopt(fstab_entry, "users") != NULL
+			  || owner);
+		if (!usable)
 			continue;
-		}
 
+		if (strcmp(fstab_entry->mnt_type, "swap") == 0)
+			continue;
+
+		mountpoint = fstab_entry->mnt_dir;
 		/* Screen out bogus mount points.  I consider "/" bogus, too. */
-		if(mountpoint[0] != '/') {
+		if (mountpoint[0] != '/' || strlen(mountpoint) < 2)
 			continue;
-		}
-		if(strlen(mountpoint) < 2) {
-			continue;
-		}
-		if(stat(mountpoint, &st1) == -1) {
-			continue;
-		}
 
-		/* Get the name of the mountpoint's parent directory. */
-		parent_dir = g_strdup(mountpoint);
-		if(strrchr(parent_dir, '/') == parent_dir) {
-			/* The mountpoint is a top-level directory. */
-			strcpy(parent_dir, "/");
-		} else {
-			/* The parent is not the root directory. */
-			g_free(parent_dir);
-			parent_dir = g_strdup_printf("%s/..", mountpoint);
-		}
-
-		/* Verify that the mountpoint and its parent exist. */
-		if((stat(mountpoint, &st1) == -1) ||
-		   (stat(parent_dir, &st2) == -1)) {
-			g_free(parent_dir);
+		res = is_mounted(mountpoint);
+		if (res == -1)
 			continue;
-		}
-		g_free(parent_dir);
 
 		/* Create a new list item and place it at the list's head. */
 		tmp = g_malloc0(sizeof(struct mountinfo));
@@ -182,7 +196,7 @@ build_mountinfo_list()
 		ret->fstype = g_strdup(fstab_entry->mnt_type);
 
 		/* Check if the filesystem is mounted. */
-		ret->mounted = (st1.st_dev != st2.st_dev);
+		ret->mounted = res == 1;
 
 		/* Check if we can write to the device.  This indicates that
 		 * we can format it and create a filesystem on it. */
@@ -191,8 +205,8 @@ build_mountinfo_list()
 		/* Now double-check that we have the software to format it. */
 		if(strcmp(ret->fstype, "auto") != 0) {
 			sbindir = g_path_get_dirname(PATH_MKFS);
-			mkfs = g_strdup_printf("%s/mkfs.%s",
-					       sbindir, ret->fstype);
+			mkfs = g_strconcat(sbindir, "/mkfs.%s", ret->fstype,
+					   NULL);
 			g_free(sbindir);
 			if(access(mkfs, X_OK) != 0) {
 				ret->writable = FALSE;
@@ -202,20 +216,22 @@ build_mountinfo_list()
 
 		/* FIXME: we make a kernel assumption which is NOT PORTABLE to
 		 * determine if it's a floppy-disk-type device. */
-		ret->fdformat = (S_ISBLK(st.st_mode) &&
-				 ((major(st.st_rdev) == 2) ||	/* fdc */
-				  (major(st.st_rdev) == 47)));	/* usb */
+		ret->fdformat = (known_device && S_ISBLK(dev_st.st_mode)
+				 && (major(dev_st.st_rdev) == 2 ||  /* fdc */
+				     major(dev_st.st_rdev) == 47)); /* usb */
 	}
 
 	endmntent(fstab);
 
+	blkid_put_cache(cache);
+
 	return ret;
 }
 
 /* Return the mountinfo struct associated with the currently-selected row in
  * the TreeView, which is stored in the fourth column in its model. */
 static struct mountinfo *
-selected_info()
+selected_info(void)
 {
 	GtkTreeModel *model;
 	GtkTreeIter iter;
@@ -229,70 +245,25 @@ selected_info()
 	return info;
 }
 
-/* Check if a given mountpoint has a filesystem mounted under it. */
-static gboolean
-check_is_mounted(struct mountinfo *info)
-{
-	char *parent_dir;
-	struct stat st1, st2;
-
-	if (info == NULL) {
-		return FALSE;
-	}
-
-	/* Screen out bogus mount points.  I consider "/" bogus, too. */
-	if (info->dir[0] != '/') {
-		return FALSE;
-	}
-	if (strlen(info->dir) < 2) {
-		return FALSE;
-	}
-	if (stat(info->dir, &st1) == -1) {
-		return FALSE;
-	}
-
-	/* Get the name of the mountpoint's parent directory. */
-	parent_dir = g_strdup(info->dir);
-	if (strrchr(parent_dir, '/') == parent_dir) {
-		/* The mountpoint is a top-level directory. */
-		strcpy(parent_dir, "/");
-	} else {
-		/* The parent is not the root directory. */
-		g_free(parent_dir);
-		parent_dir = g_strdup_printf("%s/..", info->dir);
-	}
-
-	/* Verify that the mountpoint and its parent exist. */
-	if ((stat(info->dir, &st1) == -1) ||
-	    (stat(parent_dir, &st2) == -1)) {
-		g_free(parent_dir);
-		return FALSE;
-	}
-	g_free(parent_dir);
-
-	/* Check if the filesystem is mounted. */
-	info->mounted = (st1.st_dev != st2.st_dev);
-
-	return info->mounted;
-}
-
 /* If the selection changed, we need to enable/disable specific buttons
  * depending on whether or not the selected filesystem is mounted or not. */
 static void
 changed_callback(GtkTreeSelection *ignored, gpointer also_ignored)
 {
 	struct mountinfo *info;
+
+	(void)ignored;
+	(void)also_ignored;
 	info = selected_info();
 	if (info != NULL) {
 		/* Enable/disable buttons based on whether the filesystem is
 		 * mounted or not. */
-		check_is_mounted(info);
-		gtk_widget_set_sensitive(info->format,
-					 info->writable && (!info->mounted));
-		gtk_button_set_label(GTK_BUTTON(info->mount),
-				     info->mounted ?
-				     UNMOUNT_TEXT : MOUNT_TEXT);
-		gtk_widget_set_sensitive(info->mount, TRUE);
+		info->mounted = is_mounted(info->dir) == 1;
+		gtk_widget_set_sensitive(format_button,
+					 info->writable && !info->mounted);
+		gtk_button_set_label(GTK_BUTTON(mount_button), info->mounted
+				     ? UNMOUNT_TEXT : MOUNT_TEXT);
+		gtk_widget_set_sensitive(mount_button, TRUE);
 	} else {
 		gtk_widget_set_sensitive(format_button, FALSE);
 		gtk_widget_set_sensitive(mount_button, FALSE);
@@ -302,16 +273,14 @@ changed_callback(GtkTreeSelection *ignored, gpointer also_ignored)
 static void
 format(struct mountinfo *info)
 {
-	GtkWidget *dialog, *vbox, *format;
-	GtkWidget *label, *options, *menu, *item, *table;
+	GtkWidget *dialog, *vbox, *fdformat;
+	GtkWidget *label, *options, *table;
 	GtkResponseType response;
-	char *command[5];
-	int status, defaultfs;
+	char *command[6];
 	GError *error = NULL;
-	char *sbindir = NULL, *mkfs = NULL, *fstype = NULL;
-	int i;
-	gint child;
+	char *mkfs = NULL;
 	glob_t results;
+	gboolean results_valid = FALSE;
 
 	/* Verify that the user really wants to do this, because it's
 	 * a destructive process. */
@@ -328,47 +297,41 @@ format(struct mountinfo *info)
 	/* If it's the kind of device we low-level format, let the user
 	 * disable that step. */
 	if (info->fdformat) {
-		format = gtk_check_button_new_with_mnemonic(_("Perform a "
-							      "_low-level "
-							      "format."));
-		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(format), TRUE);
-		gtk_widget_show(format);
-	} else {
-		format = NULL;
-	}
+		fdformat = gtk_check_button_new_with_mnemonic(_("Perform a "
+								"_low-level "
+								"format."));
+		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(fdformat), TRUE);
+		gtk_widget_show(fdformat);
+	} else
+		fdformat = NULL;
 
 	/* If it's a filesystem of type "auto", we also have to let the user
 	 * select a filesystem type to format it as. */
 	if (strcmp(info->fstype, "auto") == 0) {
-		sbindir = g_path_get_dirname(PATH_MKFS);
-		mkfs = g_strdup_printf("%s/mkfs.*", sbindir);
+		char *sbindir;
+		size_t i, defaultfs;
 
-		menu = gtk_menu_new();
-		gtk_widget_show(menu);
+		sbindir = g_path_get_dirname(PATH_MKFS);
+		mkfs = g_strconcat(sbindir, "/mkfs.*", NULL);
+		g_free(sbindir);
+
+		options = gtk_combo_box_new_text();
 
 		defaultfs = 0;
-		if (glob(mkfs, 0, NULL, &results) == 0) {
-			for(i = 0; i < results.gl_pathc; i++) {
-				const char *text;
-				text = results.gl_pathv[i] + strlen(mkfs) - 1;
-				item = gtk_menu_item_new_with_label(text);
-				gtk_widget_show(item);
-				if(strcmp(text, PREFERRED_FS) == 0) {
-					defaultfs = i;
-				}
-				gtk_menu_shell_append(GTK_MENU_SHELL(menu),
-						      item);
-			}
-		} else {
-			memset(&results, 0, sizeof(results));
+		results_valid = glob(mkfs, 0, NULL, &results) == 0;
+		if (!results_valid)
+			results.gl_pathc = 0;
+		for(i = 0; i < results.gl_pathc; i++) {
+			const char *text;
+
+			text = results.gl_pathv[i] + strlen(mkfs) - 1;
+			gtk_combo_box_append_text(GTK_COMBO_BOX(options),
+						  text);
+			if(strcmp(text, PREFERRED_FS) == 0)
+				defaultfs = i;
 		}
 
-		options = gtk_option_menu_new();
-		gtk_option_menu_set_menu(GTK_OPTION_MENU(options), menu);
-		if (defaultfs > 0) {
-			gtk_option_menu_set_history(GTK_OPTION_MENU(options),
-						    defaultfs);
-		}
+		gtk_combo_box_set_active(GTK_COMBO_BOX(options), defaultfs);
 		gtk_widget_show(options);
 
 		label = gtk_label_new_with_mnemonic(_("Select a _filesystem type to create:"));
@@ -380,10 +343,9 @@ format(struct mountinfo *info)
 	}
 
 	/* Run the dialog. */
-	if (format) {
-		gtk_table_attach_defaults(GTK_TABLE(table), format,
+	if (fdformat)
+		gtk_table_attach_defaults(GTK_TABLE(table), fdformat,
 					  0, 2, 0, 1);
-	}
 	if (options && label) {
 		gtk_table_attach_defaults(GTK_TABLE(table), label,
 					  0, 1, 1, 2);
@@ -391,18 +353,22 @@ format(struct mountinfo *info)
 					  1, 2, 1, 2);
 	}
 	gtk_widget_show_all(table);
-	vbox = (GTK_DIALOG(dialog))->vbox;
-	gtk_box_pack_start_defaults(GTK_BOX(vbox), table);
+	vbox = gtk_dialog_get_content_area (GTK_DIALOG(dialog));
+	gtk_box_pack_start(GTK_BOX(vbox), table, TRUE, TRUE, 0);
 
 	response = gtk_dialog_run(GTK_DIALOG(dialog));
 	gtk_widget_hide(dialog);
 
 	/* If the user said "yes", make a new filesystem on it. */
 	if (response == GTK_RESPONSE_YES) {
-		if (GTK_IS_CHECK_BUTTON(format)) {
-			GtkToggleButton *toggle = GTK_TOGGLE_BUTTON(format);
+		gint child;
+		int status;
+		char *fstype;
+
+		if (GTK_IS_CHECK_BUTTON(fdformat)) {
+			GtkToggleButton *toggle = GTK_TOGGLE_BUTTON(fdformat);
 			if (gtk_toggle_button_get_active(toggle)) {
-				command[0] = PATH_FDFORMAT;
+				command[0] = (char *)PATH_FDFORMAT;
 				command[1] = info->dev;
 				command[2] = NULL;
 				if (g_spawn_async("/",
@@ -411,56 +377,57 @@ format(struct mountinfo *info)
 						  G_SPAWN_STDOUT_TO_DEV_NULL,
 						  NULL, NULL,
 						  &child, &error)) {
+					/* FIXME: something better than busy waiting? */
 					while (waitpid(child, &status, WNOHANG) == 0) {
 						gtk_main_iteration_do(FALSE);
 					}
-					if (WIFEXITED(status)) {
-						if (WEXITSTATUS(status) == 0) {
-							info->mounted = !info->mounted;
-						}
-					}
 				}
 			}
 		}
 		fstype = NULL;
 		if (options) {
-			defaultfs = gtk_option_menu_get_history(GTK_OPTION_MENU(options));
-			if (results.gl_pathc > defaultfs) {
-				fstype = results.gl_pathv[defaultfs] + strlen(mkfs) - 1;
-			}
-			g_free(mkfs);
-			g_free(sbindir);
+			int fs;
+
+			fs = gtk_combo_box_get_active(GTK_COMBO_BOX(options));
+			if (fs >= 0 && (size_t)fs < results.gl_pathc)
+				fstype = results.gl_pathv[fs]
+					+ strlen(mkfs) - 1;
 		}
 		if (fstype == NULL) {
 			fstype = info->fstype;
 		}
-		command[0] = PATH_MKFS;
-		command[1] = "-t";
+		command[0] = (char *)PATH_MKFS;
+		command[1] = (char *)"-t";
 		command[2] = fstype;
-		command[3] = info->dev;
-		command[4] = NULL;
+		if (!strcmp(fstype,"vfat") || !strcmp(fstype,"msdos")) {
+			command[3] = (char *)"-I";
+			command[4] = info->dev;
+			command[5] = NULL;
+		} else {
+			command[3] = info->dev;
+			command[4] = NULL;
+		}
 		if (g_spawn_async("/",
 				  command, NULL,
 				  G_SPAWN_DO_NOT_REAP_CHILD |
 				  G_SPAWN_STDOUT_TO_DEV_NULL,
 				  NULL, NULL,
 				  &child, &error)) {
+			/* FIXME: something better than busy waiting? */
 			while (waitpid(child, &status, WNOHANG) == 0) {
 				gtk_main_iteration_do(FALSE);
 			}
-			if (WIFEXITED(status)) {
-				if (WEXITSTATUS(status) == 0) {
-					info->mounted = !info->mounted;
-				}
-			}
 		}
 	}
 
 	gtk_widget_destroy(dialog);
+	g_free(mkfs);
+	if (results_valid)
+		globfree(&results);
 }
 
 static void
-response_callback(GtkWidget *emitter, GtkResponseType response)
+response_callback(GtkWidget *emitter, gint response, gpointer user_data)
 {
 	struct mountinfo *info;
 	char *command[3];
@@ -469,6 +436,7 @@ response_callback(GtkWidget *emitter, GtkResponseType response)
 	GError *error = NULL;
 	GtkWidget *toplevel, *dialog;
 
+	(void)user_data;
 	info = selected_info();
 	switch (response) {
 		case GTK_RESPONSE_CLOSE:
@@ -478,18 +446,19 @@ response_callback(GtkWidget *emitter, GtkResponseType response)
 			/* Format. */
 			g_return_if_fail(info != NULL);
 			toplevel = gtk_widget_get_toplevel(emitter);
-			if (GTK_WIDGET_TOPLEVEL(toplevel)) {
+			if (gtk_widget_is_toplevel(toplevel)) {
 				gtk_widget_set_sensitive(toplevel, FALSE);
 			}
 			format(info);
-			if (GTK_WIDGET_TOPLEVEL(toplevel)) {
+			if (gtk_widget_is_toplevel(toplevel)) {
 				gtk_widget_set_sensitive(toplevel, TRUE);
 			}
 			break;
 		case ACTION_MOUNT:
 			/* Mount/unmount. */
 			g_return_if_fail(info != NULL);
-			command[0] = info->mounted ?  PATH_UMOUNT : PATH_MOUNT;
+			command[0] = info->mounted
+				? (char *)PATH_UMOUNT : (char *)PATH_MOUNT;
 			command[1] = info->dir;
 			command[2] = NULL;
 			if (g_spawn_sync("/",
@@ -502,17 +471,13 @@ response_callback(GtkWidget *emitter, GtkResponseType response)
 					 &messages,
 					 &status,
 					 &error)) {
-				if (WIFEXITED(status)) {
-					if (WEXITSTATUS(status) == 0) {
-						info->mounted = !info->mounted;
+				if (messages != NULL) {
+					if (strlen(messages) > 0) {
+						dialog = create_error_box(messages,
+									  NULL);
+						gtk_dialog_run(GTK_DIALOG(dialog));
+						gtk_widget_destroy(dialog);
 					}
-				}
-				if ((messages != NULL) &&
-				    (strlen(messages) > 0)) {
-					dialog = create_error_box(messages,
-								  NULL);
-					gtk_dialog_run(GTK_DIALOG(dialog));
-					gtk_widget_destroy(dialog);
 					g_free(messages);
 				}
 			}
@@ -523,10 +488,10 @@ response_callback(GtkWidget *emitter, GtkResponseType response)
 	}
 }
 
-static void
-create_usermount_window()
+static gboolean
+create_usermount_window(void)
 {
-	GtkWidget *dialog, *treeview, *format, *mount;
+	GtkWidget *dialog, *treeview;
 	GtkTreeIter iter;
 	GtkListStore *model;
 	GtkTreeViewColumn *column[3];
@@ -542,12 +507,13 @@ create_usermount_window()
 		dialog = create_error_box(_("There are no filesystems which you are allowed to mount or unmount.\nContact your administrator."),
 					  _("User Mount Tool"));
 		g_signal_connect_object(G_OBJECT(dialog), "destroy",
-					GTK_SIGNAL_FUNC(gtk_main_quit), NULL,
+					G_CALLBACK(gtk_main_quit), NULL,
 					G_CONNECT_AFTER);
 		g_signal_connect_object(G_OBJECT(dialog), "delete-event",
-					GTK_SIGNAL_FUNC(gtk_main_quit), NULL,
+					G_CALLBACK(gtk_main_quit), NULL,
 					G_CONNECT_AFTER);
-		return;
+		gtk_dialog_run(GTK_DIALOG(dialog));
+		return FALSE;
 	}
 
 	/* Create the dialog. */
@@ -559,24 +525,30 @@ create_usermount_window()
 					     NULL);
 
 	/* Connect default signal handlers. */
-	g_signal_connect(G_OBJECT(dialog), "destroy",
-			 GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
+	g_signal_connect(G_OBJECT(dialog), "destroy", G_CALLBACK(gtk_main_quit),
+			 NULL);
 	g_signal_connect(G_OBJECT(dialog), "delete-event",
-			 GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
+			 G_CALLBACK(gtk_main_quit), NULL);
+
+	/* Set the window icon */
+	gtk_window_set_icon_from_file(GTK_WINDOW(dialog),
+				      PIXMAPDIR "/disks.png", NULL);
 
 	/* Create the other buttons. */
-	format_button = format = gtk_button_new_with_mnemonic(_("_Format"));
-	gtk_widget_set_sensitive(format, FALSE);
-	gtk_dialog_add_action_widget(GTK_DIALOG(dialog), format, ACTION_FORMAT);
-	mount_button = mount = gtk_button_new_with_mnemonic(info->mounted ?
-							    UNMOUNT_TEXT :
-							    MOUNT_TEXT);
-	gtk_widget_set_sensitive(mount, FALSE);
-	gtk_dialog_add_action_widget(GTK_DIALOG(dialog), mount, ACTION_MOUNT);
+	format_button = gtk_button_new_with_mnemonic(_("_Format"));
+	gtk_widget_set_sensitive(format_button, FALSE);
+	gtk_dialog_add_action_widget(GTK_DIALOG(dialog), format_button,
+				     ACTION_FORMAT);
+	mount_button = gtk_button_new_with_mnemonic(info->mounted
+						    ? UNMOUNT_TEXT
+						    : MOUNT_TEXT);
+	gtk_widget_set_sensitive(mount_button, FALSE);
+	gtk_dialog_add_action_widget(GTK_DIALOG(dialog), mount_button,
+				     ACTION_MOUNT);
 
 	/* Set a signal handler to handle button presses. */
 	g_signal_connect(G_OBJECT(dialog), "response",
-			 GTK_SIGNAL_FUNC(response_callback), NULL);
+			 G_CALLBACK(response_callback), NULL);
 
 	/* Build a tree model with this data. */
 	model = gtk_list_store_new(4,
@@ -587,8 +559,6 @@ create_usermount_window()
 
 	/* Now add a row to the list using the data. */
 	for (i = info; i != NULL; i = i->next) {
-		i->format = format;
-		i->mount = mount;
 		gtk_list_store_append(model, &iter);
 		gtk_list_store_set(model, &iter,
 				   0, i->dev,
@@ -628,25 +598,68 @@ create_usermount_window()
 	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
 	tree_selection = selection;
 	g_signal_connect(G_OBJECT(selection), "changed",
-		         GTK_SIGNAL_FUNC(changed_callback), NULL);
+			 G_CALLBACK(changed_callback), NULL);
 
 	/* Pack it in and show the dialog. */
-	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), treeview,
-			   TRUE, TRUE, 0);
+	gtk_box_pack_start(GTK_BOX(gtk_dialog_get_content_area
+				   (GTK_DIALOG(dialog))),
+			   treeview, TRUE, TRUE, 0);
 
 	gtk_widget_show_all(dialog);
+
+	return TRUE;
+}
+
+static void
+format_one_device (const char *device)
+{
+	struct mountinfo *info, *i;
+
+	info = build_mountinfo_list();
+
+	for (i = info; i != NULL; i = i->next) {
+		if (strcmp (i->dev, device) == 0)
+			break;		
+	}
+
+	if (i && i->writable && !i->mounted) {
+		format (i);
+	} else {	
+		GtkWidget *dialog;
+
+		dialog = gtk_message_dialog_new (NULL, 
+						 0,
+						 GTK_MESSAGE_ERROR,
+						 GTK_BUTTONS_OK,
+						 _("The device '%s' can not be formatted."),
+						 device);
+                gtk_dialog_run(GTK_DIALOG(dialog));
+	}
 }
 
 int
 main(int argc, char *argv[])
 {
-	bindtextdomain(PACKAGE, DATADIR "/locale");
+	bindtextdomain(PACKAGE, LOCALEDIR);
 	bind_textdomain_codeset(PACKAGE, "UTF-8");
 	textdomain(PACKAGE);
-	gtk_set_locale();
 	gtk_init(&argc, &argv);
-	glade_init();
-	create_usermount_window();
-	gtk_main();
+
+        if (argc > 1) {
+		char *name = strrchr (argv[0], '/');
+		
+		name = name ? name + 1: argv[0];
+		if ( !strcmp (name, "userformat") ) {
+        		format_one_device (argv[1]);
+		}
+        } else {
+		gboolean run_main;
+
+		run_main = create_usermount_window();
+		if (run_main) {
+			gtk_main();
+		}
+	}
+
 	return 0;
 }
